#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <ls-b247\lsapi\lsapi.h>
#include "lsb_contract.h"
#include "_lsb_dbghelp.h"

// ASSERTs, etc. are not allowed in this file! 
// That could cause infinite recursion.        


#ifdef USE_DBGHELP

DWORD LSBASE_NAMESPACE::g_GrabbedContext[sizeof(CONTEXT)/sizeof(DWORD)];

StackTrace::StackTrace(CONTEXT &cx_)
  : cx(cx_), hProcess(GetCurrentProcess()), bDeinit(false)
{
  ZeroMemory(&sf, sizeof(sf));
  sf.AddrPC.Offset = cx.Eip;
  sf.AddrPC.Mode = AddrModeFlat;
  sf.AddrFrame.Offset = cx.Ebp;
  sf.AddrFrame.Mode = AddrModeFlat;
  sf.AddrStack.Offset = cx.Esp;
  sf.AddrStack.Mode = AddrModeFlat;

  ZeroMemory(&module, sizeof(module));
  module.SizeOfStruct = sizeof(module); //0x248;

  ZeroMemory(&sym, sizeof(symbol_buffer));
  sym.SizeOfStruct = sizeof(symbol_buffer);
  sym.MaxNameLength = sizeof(symbol_buffer)-sizeof(sym);

  ZeroMemory(&line, sizeof(line));
  line.SizeOfStruct = sizeof(line);

  SymSetOptions(
    SYMOPT_LOAD_LINES|      /* we want line numbers */ 
    SYMOPT_DEFERRED_LOADS|  /* only load symbols when needed */ 
    SYMOPT_UNDNAME|         /* undecorate the names (like source) */ 
    SYMOPT_FAIL_CRITICAL_ERRORS); /* don't whine if drive letters are wrong */ 
}


void StackTrace::GetCurrentPosition(char *buffer, unsigned long size)
{
  LPSTR values[7];
  if (!GetModule())
  {
    values[0] = (LPSTR)sf.AddrPC.Offset;
    FormatMessage(
      FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
      "%1!08X! (module unknown)",
      NULL,NULL,buffer,size,values);
    return;
  }
  if (!GetSymbol())
  {
    values[0] = (LPSTR)sf.AddrPC.Offset;
    values[1] = (LPSTR)module.ModuleName; /* or ImageName */ 
    values[2] = (LPSTR)(sf.AddrPC.Offset - module.BaseOfImage);
    FormatMessage(
      FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
      "%2 + %3!08X!(%3!d!) = %1!08X!",
      NULL,NULL,buffer,size,values);
    return;
  }
  if (!GetLine())
  {
    values[0] = (LPSTR)sf.AddrPC.Offset;
    values[1] = (LPSTR)module.ModuleName; /* or ImageName */ 
    values[2] = (LPSTR)(sf.AddrPC.Offset - module.BaseOfImage);
    values[3] = (LPSTR)sym.Name;
    values[4] = (LPSTR)(sf.AddrPC.Offset - sym.Address);
    FormatMessage(
      FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
      "%2 + %3!08X!(%3!d!) = %4 + %5!d! bytes = %1!08X!",
      NULL,NULL,buffer,size,values);
    return;
  }
  values[0] = (LPSTR)sf.AddrPC.Offset;
  values[1] = (LPSTR)module.ModuleName; /* or ImageName */ 
  values[2] = (LPSTR)(sf.AddrPC.Offset - module.BaseOfImage);
  values[3] = (LPSTR)sym.Name;
  values[4] = (LPSTR)(sf.AddrPC.Offset - sym.Address);
  values[5] = (LPSTR)line.FileName;
  values[6] = (LPSTR)line.LineNumber;
  FormatMessage(
    FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
    "%6 (%7!d!) : in %4 + %5!d! bytes = %1!08X!",
    NULL,NULL,buffer,size,values);
}

#endif

