#include "module.h"
#include <tchar.h>
#define atoi not_atoi
#include <stdlib.h>
// use one-at-a-time hash for keys
#define GIMME_HASH

#define OPEN_CHAR _T('{')
#define CLOSE_CHAR _T('}')
#define ALT_OPEN_CHAR _T('[')
#define ALT_CLOSE_CHAR _T(']')
#define ESCAPE_CHAR _T('\\')
#define LOCAL_CHAR _T('@')
#define BANG_CHAR _T('!')
#define PARAM_OPEN_CHAR _T('(')
#define PARAM_CLOSE_CHAR _T(')')
#define ERROR_SIGNAL_CHAR _T('#')
#define RESERVED_PREFIX_CHARS _T("\\!@#$%^&*-=+`~;:,.<>/?|")

// you need to specify ESCAPE_CHAR if you want it escaped
// escape everything: \\ 
#define ESCAPE_ALL_CHAR ESCAPE_CHAR
#define ESCAPED_CHARS _T("\\'^$\n\"\r;\t{}[]")
#define ESCAPED_TO    _T("\\acdnqrst{}[]")
// escape only what's needed for label: \| 
#define ESCAPE_LABEL_CHAR _T('|')
#define ESCAPED_LABEL_CHARS _T("\\[]\n")
#define ESCAPED_LABEL_TO    _T("\\[]n")

#define SHELL_VAR_START _T('$')
#define SHELL_VAR_END _T('$')

#ifdef GIMME_HASH
#include "hash-fnv.h"
typedef hasher::fnv designated_hash_function;
#endif

inline int _wctot(TCHAR *dest, wchar_t src)
#ifndef UNICODE
{ return wctomb(dest,src); }
#else
{ *dest = src; return 1; }
#endif

/*
  {} -> local default string
  {@aa} -> local var @aa
  {aa} {_aa} -> global var aa (for now becomes $aa$)
  {!aa} -> exec !aa, replace with string passed to !result
  {$aa} {"aa} {'aa} -> should be impossible
  {1aa}-{9aa} -> special, reserved

  {^aa} {#aa} {&aa} {%aa} {*aa} {~aa} {:aa} {.aa}
  {,aa} {<aa} {?aa} {/aa} {=aa} {-aa} {+aa} {|aa}
  {>aa} {`aa} {0aa}
   -> all reserved

  \[ -> [
  \] -> ]
  \{ -> {
  \} -> }
  \\ -> \
  \1x -> \x
  \2x -> \1x
  \10x -> \9x
  \200x -> \199x
  {{{\}}} -> {}} (process \, end at matching '}}')
  {[{{\}]} -> {{\} (ignore \, end at first ']}')
  {1{...}} -> {{...}}
  {2{...}} -> {1{...}}
  \n -> newline
  \d -> dollar
  \c -> caret
  \q -> quote
  \a -> apostrophe
  \x00 -> hex code 0x00
  \u0000 -> unicode ?

  {abc\\} -> expand abc, escape everything
  {abc\|} -> expand abc, escape stuff for label
  {abs\Name} -> expand abc, apply TransformName (esc, tr)

  {=abc} -> evaluate abc an extra time? (needs to allow nesting {}'s)


0: \103d\1d\201d\5d\12d : [2[quote\d[1[moo]]\d]] : [4{literal\d\[{moo}]\d}]
1: \102d\d\200d\4d\11d : [1[quote\d[1[moo]]\d]] : [3{literal\d\[{moo}]$}]
2: \101d$\199d\3d\10d : [[quote\d[1[moo]]\d]] : [2{literal\d\[{moo}]$}]
3: \100d$\198d\2d\9d : quote$[1[moo]]$ : [1{literal\d\[{moo}]$}]
4: \99d$\197d\1d\8d : quote$[[moo]]$ : [{literal\d\[{moo}]$}]
5: \98d$\196d\d\7d : quote$moo$ : literal\d\[{moo$}]
6: \97d$\195d$\6d : quote$moo$ : literal$[{moo$}]


*/


#define PAIR(a,b) ((((DWORD)(a))<<16)|((DWORD)(b)))
#define DEST_INC(a) ((a)>>16)
#define SRC_INC(a) ((a)&0xffff)

#define ONECHAR(c) if (true) { *dest = _T(c); return PAIR(1,1); } else
#define NOCHAR(b) return PAIR(0,(b))

inline static DWORD xdigit2int(TCHAR c)
{
  ASSERT((c>=_T('0') && c<=_T('9'))||(c>=_T('a')&&c<=_T('f'))||(c>=_T('A')&&c<=_T('F')));
  return (c&15)+((c&64)?10:0);
}

// rcVarCopyRep: copy and decrement an escape depth number
// returns true on success, false on failure
// this does not check *src, it must be '1'-'9'.
// src, dest, maxchars are adjusted to point after the stuff
// first is inserted before everything
// if second is not 0, the char after the numbers must match it
// the number is handled as a string, so there's no practical limit
static bool rcVarCopyRep(TCHAR *&src, TCHAR *&dest, size_t &maxchars, TCHAR first, TCHAR second)
{
  bool b;
  TCHAR *ptr;
  DWORD v;
  int i;

  REQUIRE_PTR(src, READABLE(src));
  REQUIRE_PTR(dest, WRITEABLEN(dest,maxchars));
  REQUIRE(*src>=_T('1') && *src<=_T('9'));

  b = (*src==_T('1'));
  ptr = src+1;
  /* find the end of the number */ 
  while (*ptr >= _T('0') && *ptr <= _T('9'))
    { b = b && (*ptr==_T('0')); ++ptr; }

  /* check the dest. length */ 
  v = ptr-src + (b?0:1);
  if (v+1 > maxchars) return false;

  /* check the following char */ 
  if (second!=0 && second!=*ptr) return false;

  /* we're gonna keep the prec char, so write it */ 
  *dest++ = first;
  --maxchars;

  if (b)
  {
    /* if the number is 1, omit it */ 
    if (v==1) 
    {
      *dest++ = *ptr++;
      --maxchars;
      src = ptr;
    }
    else
    {
      /* otherwise it'll be all 9's */ 
      src = ptr;
      while (--v) { *dest++ = _T('9'); --maxchars; }
      *dest++ = *src++;
      --maxchars;
    }
  }
  else
  {
    /* copy */ 
    while (src<=ptr) { *dest++ = *src++; --maxchars; }
    /* decrement */ 
    for (i=2;(DWORD)i<=v;++i)
    {
      if (dest[-i] == _T('0')) dest[-i] = _T('9');
      else { --dest[-i]; break; }
    }
  }
  ENSURE((signed)maxchars>0);
  return true;
}

// maxchars must be at least 2 (can take one char and null)
// returns PAIR(dest chars inserted, source chars used)
static DWORD rcVarEscape(TCHAR *src, TCHAR *dest, size_t maxchars)
{
  REQUIRE(maxchars>1);
  REQUIRE_PTR(src, READABLE(src));
  REQUIRE_PTR(dest, WRITEABLEN(dest,maxchars));

  DWORD v;
  bool b=false;
  TCHAR c = *src;
  TCHAR *ptr;

  switch (c)
  {
  /* after we know what the char is,  */ 
  /* we can use ++ instead of _tcsinc */ 
  case OPEN_CHAR:
  case CLOSE_CHAR:
  case ALT_OPEN_CHAR:
  case ALT_CLOSE_CHAR:
  case ESCAPE_CHAR: ONECHAR(c);
  case _T('a'): case _T('A'): ONECHAR('\'');
  case _T('c'): case _T('C'): ONECHAR('^');
  case _T('d'): case _T('D'): ONECHAR('$');
  case _T('n'): case _T('N'): ONECHAR('\n');
  case _T('q'): case _T('Q'): ONECHAR('"');
  case _T('r'): case _T('R'): ONECHAR('\r');
  case _T('s'): case _T('S'): ONECHAR(';');
  case _T('t'): case _T('T'): ONECHAR('\t');

  case _T('x'): case _T('X'): /* \x00 - \xFF */ 
    if (!_istxdigit(*++src)) NOCHAR(3);
    *dest = static_cast<TCHAR>(xdigit2int(*src)<<4);
    if (!_istxdigit(*++src)) NOCHAR(3);
    *dest += static_cast<TCHAR>(xdigit2int(*src));
    return PAIR(1,3);

  case _T('u'): case _T('U'): /* \u0000 - \xFFFF */ 
    /* might be double-byte char */ 
    if (maxchars<3) NOCHAR(5);
    if (!_istxdigit(*++src)) NOCHAR(5);
    v = xdigit2int(*src)<<12;
    if (!_istxdigit(*++src)) NOCHAR(5);
    v += xdigit2int(*src)<<8;
    if (!_istxdigit(*++src)) NOCHAR(5);
    v += xdigit2int(*src)<<4;
    if (!_istxdigit(*++src)) NOCHAR(5);
    v += xdigit2int(*src);
    v = _wctot(dest, static_cast<wchar_t>(v));
    if (v<0) NOCHAR(5);
    return PAIR(v,5);

  case _T('1'): b = true; /* may need to have 1 fewer digit */ 
  case _T('2'): case _T('3'): case _T('4'): case _T('5'):
  case _T('6'): case _T('7'): case _T('8'): case _T('9'): /* note: no '0'! */ 
    ptr = src;
    v = maxchars;
    if (!rcVarCopyRep(ptr,dest,maxchars,ESCAPE_CHAR,0))
      NOCHAR(ptr-src);
    ASSERT(maxchars<v && ((int)maxchars)>0);
    ASSERT(ptr>src);
    return PAIR(v-maxchars,ptr-src);

  default: NOCHAR(1);
    /* should call user-defined stuff so I can use /; as a fieldsep */ 
  }
}

enum {
  vtError,
  vtReserved,
  vtPlain,/* {abcd} */ 
  vtAt,/* {@abc} */ 
  vtBang,/* {!abc} */ 
  vtDefault,/* {} */ 
  vtfSpecialBlock=0x80,
  vtQuote,/* {{ab}} */ 
  vtLiteral /* {]ab]} */ 
};

// figure out what type of {} block this is
static int rcVarType(TCHAR *src)
{
  bool b=false;

  ASSERT_READABLE(src);

  /* forms with special bracket matching */ 
  switch (*src)
  {
  case _T('1'): b = true; /* may need to have 1 fewer digit */ 
  case _T('2'): case _T('3'): case _T('4'): case _T('5'):
  case _T('6'): case _T('7'): case _T('8'): case _T('9'): /* note: no '0'! */ 
    while (_istdigit(*src)) ++src;
    if (*src==OPEN_CHAR) return vtQuote;
    if (*src==ALT_OPEN_CHAR) return vtLiteral;
    return vtError;

  case OPEN_CHAR:
    return vtQuote;
    /* {{ end at matching }} */ 

  case ALT_OPEN_CHAR:
    return vtLiteral;
    /* {[ end at first ]} */ 
  }

  /* skip whitespace */ 
  while (_istspace(*src)) ++src;

  ASSERT_READABLE(src);

  /* check name for type */ 
  switch (*src)
  {
  case PARAM_CLOSE_CHAR: /* what is this supposed to be? */ 
  case ALT_CLOSE_CHAR:
  case OPEN_CHAR: /* { { } } is bad form, it must be {{ }} */ 
  case ALT_OPEN_CHAR: return vtError;

  case ESCAPE_CHAR:
  case CLOSE_CHAR: /* local default value {} */ 
  case PARAM_OPEN_CHAR: return vtDefault;

  case _T('^'): case _T('#'): case _T('&'): case _T('%'): case _T('*'):
  case _T('~'): case _T(':'): case _T('.'): case _T(','): case _T('<'):
  case _T('?'): case _T('/'): case _T('='): case _T('-'): case _T('+'):
  case _T('|'): case _T('>'): case _T('`'): case _T('0'):
    return vtReserved;

  case LOCAL_CHAR: return vtAt;
  case BANG_CHAR: return vtBang;
  default: return vtPlain;
  }
}

// evaluate intelligent quote
// handles {2{...}}, {{...}}
// {{ ends at matching }}, all {} inside must match
// when the {{}}'s are removed, \escapes are evaluated, vars are not
// escapes in nested {}'s are not evaluated
static DWORD rcVarDoQuote(TCHAR *src, TCHAR *dest, size_t maxchars, TCHAR *&outside)
{
  DWORD v;
  unsigned nDepth = 1;
  bool bDelay = false;
  TCHAR *od = dest;

  ASSERT_READABLE(src);
  ASSERT_WRITEABLEC(dest,maxchars);

  if (*src>=_T('1') && *src<=_T('9'))
  {
    if (!rcVarCopyRep(src,dest,maxchars,OPEN_CHAR,OPEN_CHAR))
      return 0;
    bDelay = true;
  }
  else
  {
    ASSERT(*src == OPEN_CHAR);
    ++src;
  }

  while (*src && maxchars && nDepth)
  {
    switch (*src)
    {
    case OPEN_CHAR:
      ++nDepth;
      break;

    case CLOSE_CHAR:
      --nDepth;
      break;

    case ESCAPE_CHAR:
      if (!bDelay && nDepth==1)
      {
        v = rcVarEscape(++src, dest, maxchars);
        ASSERT(DEST_INC(v)<maxchars);
        dest += DEST_INC(v);
        maxchars -= DEST_INC(v);
        v = SRC_INC(v);
        /* don't let skips go over the end */ 
        while (v--) if (!*++src) break;
        continue;
      }
      break;
    }
    _tccpy(dest, src);
    dest = _tcsinc(dest);
    src = _tcsinc(src);
    --maxchars;
  }
  outside = src;
  /* couldn't find the end */ 
  if (nDepth>0 || *src!=CLOSE_CHAR) return 0;

  ++outside;
  /* we have one CLOSE_CHAR in dest, remove it or add the second */ 
  if (bDelay)
  {
    if (maxchars<2) return 0;
    *dest++ = CLOSE_CHAR;
  }
  else
  {
    --dest;
  }
  return dest-od;
}

// evaluate a literal quote
// handles {2[...]}, {[...]}
// {[ ends at matching ]}, but other {} inside are not matched
// when {[]}'s are removed, NOTHING is evaluated
static DWORD rcVarDoLiteral(TCHAR *src, TCHAR *dest, size_t maxchars, TCHAR *&outside)
{
  TCHAR *od = dest;
  bool bDelay = false;
  bool bHalfOpen = false;
  bool bHalfClosed = false;
  unsigned nDepth = 1;

  ASSERT_READABLE(src);
  ASSERT_WRITEABLEC(dest,maxchars);

  if (*src>=_T('1') && *src<=_T('9'))
  {
    if (!rcVarCopyRep(src,dest,maxchars,OPEN_CHAR,ALT_OPEN_CHAR))
      return 0;
    ASSERT_READABLE(src);
    ASSERT_WRITEABLEC(dest,maxchars);
    bDelay = true;
  }
  else
  {
    ASSERT(*src == ALT_OPEN_CHAR);
    ++src;
  }

  while (*src && maxchars && nDepth)
  {
    switch (*src)
    {
    case ALT_CLOSE_CHAR:
      bHalfClosed = true;
      bHalfOpen = false;
      break;

    case CLOSE_CHAR:
      if (bHalfClosed) --nDepth;
      bHalfOpen = bHalfClosed = false;
      break;

    case OPEN_CHAR:
      bHalfClosed = false;
      bHalfOpen = !bHalfOpen;
      break;

    case ALT_OPEN_CHAR:
      if (bHalfOpen)
        ++nDepth;
      bHalfOpen = bHalfClosed = false;
      break;

    default:
      bHalfOpen = bHalfClosed = false;
    }

    ASSERT(!(bHalfOpen && bHalfClosed));
    _tccpy(dest, src);
    dest = _tcsinc(dest);
    src = _tcsinc(src);
    --maxchars;
  }

  outside = src;
  /* couldn't find end */ 
  if (nDepth>0) return 0;
  /* remove the end chars */ 
  if (!bDelay) dest -= 2;
  /* note: empty quote string will fail */ 
  return dest-od;
}

// find the matching bracket for a regular var; nesting is not allowed
static bool rcFindMatchingBracket(TCHAR *src, rcVarPointers &ptrs, TCHAR *&outside)
{
  unsigned nDepth = 1;
  bool bStartFound = false;
  ASSERT_READABLE(src);
  ASSERT_WRITEABLE(&ptrs);

  ptrs.VarStart = src;
  while (nDepth)
  {
    switch (*src)
    {
    case _T(' '): /* skip leading whitespace */ 
    case _T('\n'):
    case _T('\t'):
      if (!bStartFound)
        ptrs.VarStart++;
      break;

    case OPEN_CHAR:
      bStartFound = true;
      ++nDepth;
      break;

    case CLOSE_CHAR:
      bStartFound = true;
      --nDepth;
      break;

    case 0:
      /* no end */ 
      ptrs.VarStart = NULL;
      return false;

    default:
      bStartFound = true;
    }
    ++src;
  }
  /* trailing whitespace is cropped in rcParseVarBits */ 
  ptrs.VarEnd = src-1;
  outside = src;
  return true;
}

// parse out and chop up the various pieces: {name(params)\transform}
static void rcParseVarBits(rcVarPointers &ptrs)
{
  TCHAR *tail;
#ifdef GIMME_HASH
  designated_hash_function hash;
#endif

  if (!ptrs.VarStart) return;
  ASSERT_READABLE(ptrs.VarStart);
  ASSERT_WRITEABLE(ptrs.VarEnd);
  ptrs.VarEndChar = *ptrs.VarEnd; *ptrs.VarEnd = 0;

  /* find the end of the first token */ 
  tail = ptrs.NameStart = ptrs.VarStart;
  while (*tail
    && !_istspace(*tail)
    && *tail != PARAM_OPEN_CHAR
    && *tail != ESCAPE_CHAR)
  {
    if (_tcschr(RESERVED_PREFIX_CHARS,*tail)==0)
#ifdef GIMME_HASH
      hash(static_cast<TCHAR>(_totlower(*tail)))
#endif
      ;
    else
      if (tail!=ptrs.NameStart) break;
    ++tail;
  }
  ptrs.NameEnd = tail;
#ifdef GIMME_HASH
  hash.finish();
  ptrs.NameHash = hash;
#else
  ptrs.NameHash = 0;
#endif

  /* look for parameters */ 
  ptrs.ParamStart = _tcschr(tail, PARAM_OPEN_CHAR);
  if (ptrs.ParamStart)
  {
    ptrs.ParamEnd = _tcschr(++ptrs.ParamStart, PARAM_CLOSE_CHAR);

    if (!ptrs.ParamEnd)
      ptrs.ParamStart = NULL;
    else
      tail = ptrs.ParamEnd+1;
  }

  /* look for transform */ 
  ptrs.TransformStart = _tcschr(tail, ESCAPE_CHAR);
#ifdef GIMME_HASH
  hash.reset();
#endif
  if (ptrs.TransformStart)
  {
    ptrs.TransformEnd = ++ptrs.TransformStart;
    while (*ptrs.TransformEnd
      && !_istspace(*ptrs.TransformEnd))
    {
      if (_tcschr(RESERVED_PREFIX_CHARS,*ptrs.TransformEnd)==0)
#ifdef GIMME_HASH
        hash(static_cast<TCHAR>(_totlower(*ptrs.TransformEnd)))
#endif
        ;
      else
        if (ptrs.TransformEnd!=ptrs.TransformStart+1) break;
      ++ptrs.TransformEnd;
    }
    tail = ptrs.TransformEnd;
  }
#ifdef GIMME_HASH
  hash.finish();
  ptrs.TransformHash = hash;
#else
  ptrs.TransformHash = 0;
#endif

  /* isolate the bits */ 
  ASSERT_WRITEABLE(ptrs.NameEnd);
  ptrs.NameEndChar = *ptrs.NameEnd; *ptrs.NameEnd = 0;
  if (ptrs.ParamStart) { ptrs.ParamEndChar = *ptrs.ParamEnd; *ptrs.ParamEnd = 0; }
  if (ptrs.TransformStart) { ptrs.TransformEndChar = *ptrs.TransformEnd; *ptrs.TransformEnd = 0; }
}

// replace the chars we zeroed out to chop up the var bits
static void rcRepairVarBits(rcVarPointers &ptrs)
{
  ASSERT_WRITEABLE(&ptrs);
  ASSERT_WRITEABLE(ptrs.NameEnd);
  ASSERT_WRITEABLE(ptrs.VarEnd);

  if (ptrs.TransformStart) *ptrs.TransformEnd = ptrs.TransformEndChar;
  if (ptrs.ParamStart) *ptrs.ParamEnd = ptrs.ParamEndChar;
  *ptrs.NameEnd = ptrs.NameEndChar;
  *ptrs.VarEnd = ptrs.VarEndChar;
}

// make sure appropriate cleanup is done
class BeginVarBlock
{
protected:
  rcVarPointers &ptrs;
public:
  BeginVarBlock(rcVarPointers &p) : ptrs(p) { rcParseVarBits(ptrs); }
  ~BeginVarBlock() { rcRepairVarBits(ptrs); }
};


// {abc} -> do a VarExpansion from $abc$
static size_t rcVarStdPlainLookup(rcVarPointers *var, TCHAR *dest, size_t maxchars, DWORD)
{
  const static TCHAR VarEndString[] = { SHELL_VAR_END, 0 };
  TCHAR result[MAX_LINE_LENGTH];
  TCHAR expr[MAX_LINE_LENGTH];
  size_t len;

  ASSERT_WRITEABLE(var);
  ASSERT_WRITEABLEC(dest,maxchars);
  ASSERT_READABLE(var->NameStart);
  ASSERT((var->NameEnd-var->NameStart+2)<MAX_LINE_LENGTH);

  *expr = SHELL_VAR_START;
  _tcscpy(expr+1,var->NameStart);
  _tcscat(expr,VarEndString);
  *result = 0;
  VarExpansion(result,expr);

  /* want length in TCHARs, not # of characters */ 
  len = strlen(result); /* so not _tcslen() */ 
  if (len>=maxchars) return 0;
  _tcscpy(dest, result);
  return len;
}

// replace chars matching those in esc with a backslash and the matching char from rep
// if rep is shorter than esc, the remaining chars are just preceded by a backslash
static size_t rcTransformEscape(TCHAR *esc, TCHAR *rep, TCHAR *src, size_t slen, TCHAR *dest, size_t dsize)
{
  TCHAR *ptr;
  size_t od = dsize;
  unsigned repcnt,index;
  ASSERT_READABLE(esc);
  ASSERT_READABLE(rep);
  ASSERT_READABLEC(src,slen);
  ASSERT_WRITEABLEC(dest,dsize);

  ptr = rep;
  while (*ptr) ++ptr;
  repcnt = ptr-rep;

  while (slen--)
  {
    /* should we escape this char? */ 
    ptr = _tcschr(esc,*src);
    if (ptr!=0)
    {
      /* yes, write the escape char */ 
      if (!--dsize) return 0;
      *dest++ = ESCAPE_CHAR;
      /* then the appropriate code */ 
      if (!--dsize) return 0;
      index = (ptr-esc);
      if (index<repcnt)
        *dest++ = *((ptr-esc)+rep);
      else
        *dest++ = *src;
      /* next source char */ 
      ++src;
    }
    else
    {
      /* write the char */ 
      if (!--dsize) return 0;
      *dest++ = *src++;
    }
  }

  return od-dsize;
}

// replace chars matching those in rem with the corresponding ones in rep
// if rem is longer than rep, those matching extras in rem are discarded
static size_t rcTransformTranslate(TCHAR *rem, TCHAR *rep, TCHAR *src, size_t slen, TCHAR *dest, size_t dsize)
{
  TCHAR *ptr;
  size_t od = dsize;
  unsigned repcnt,index;
  ASSERT_READABLE(rem);
  ASSERT_READABLE(rep);
  ASSERT_READABLEC(src,slen);
  ASSERT_WRITEABLEC(dest,dsize);

  ptr = rep;
  while (*ptr) ++ptr;
  repcnt = ptr-rep;

  while (slen--)
  {
    /* should we translate this char? */ 
    ptr = _tcschr(rem,*src);
    if (ptr!=0)
    {
      index = (ptr-rem);
      if (index<repcnt)
      {
        /* write the appropriate replacement */ 
        if (!--dsize) return 0;
        *dest++ = *((ptr-rem)+rep);
      }
      /* next source char */ 
      ++src;
    }
    else
    {
      /* write the original char */ 
      if (!--dsize) return 0;
      *dest++ = *src++;
    }
  }

  return od-dsize;
}

static size_t rcVarApplyTransform(rcVarPointers *var, TCHAR *dest, size_t maxchars, size_t writtenchars, DWORD)
{
  enum {
    tNormal, tBracket, tAll
  } SpecialTransform = tNormal;
  TCHAR TransformLine[MAX_LINE_LENGTH];
  const TCHAR *rest;
  TCHAR token1[MAX_LINE_LENGTH];
  TCHAR token2[MAX_LINE_LENGTH];
  size_t len;

  ASSERT_WRITEABLE(var);
  ASSERT_WRITEABLEC(dest,maxchars);
  ASSERT(writtenchars<=maxchars);

  /* no transform */ 
  if (!var->TransformStart) return writtenchars;

  ASSERT_READABLE(var->TransformStart);
  TRACE_MSG(("Transform: %s", var->TransformStart));
  switch (*var->TransformStart)
  {
  case ESCAPE_LABEL_CHAR: SpecialTransform = tBracket; break;
  case ESCAPE_ALL_CHAR: SpecialTransform = tAll; break;
  }
  if (_tcschr(RESERVED_PREFIX_CHARS,*var->TransformStart)!=NULL)
    ++var->TransformStart;

  *TransformLine = 0;
  if (*var->TransformStart)
  {
    _tcscpy(token1, _T("Transform"));
    _tcscpy(token1+9, var->TransformStart);
    GetRCLine(token1, TransformLine, MAX_LINE_LENGTH, _T(""));
  }

  if (SpecialTransform != tNormal)
  {
    switch (SpecialTransform)
    {
    case tAll:
      len = rcTransformEscape(ESCAPED_CHARS,ESCAPED_TO,dest,writtenchars,token1,MAX_LINE_LENGTH);
      break;

    case tBracket:
      len = rcTransformEscape(ESCAPED_LABEL_CHARS,ESCAPED_LABEL_TO,dest,writtenchars,token1,MAX_LINE_LENGTH);
      break;
    }
    if (!len) return 0;
    if (len>maxchars) return 0;
    token1[len] = 0;
    _tcscpy(dest,token1);
    writtenchars = len;
  }

  if (*TransformLine)
  do /* provide a block to break; out of */ 
  {
    len = 0;
    if (!GetToken(TransformLine, token1, &rest, FALSE))
      break;
    /* user-defined escape transform */ 
    if (_tcsicmp(token1, _T("esc"))==0)
    {
      if (!GetToken(rest, token1, &rest, FALSE)) break;
      if (!GetToken(rest, token2, &rest, FALSE)) *token2 = 0;
      len = rcTransformEscape(token1, token2, dest, writtenchars, TransformLine, MAX_LINE_LENGTH);
    }
    /* user-defined translate transform */ 
    else if (_tcsicmp(token1, _T("tr"))==0)
    {
      if (!GetToken(rest, token1, &rest, FALSE)) break;
      if (!GetToken(rest, token2, &rest, FALSE)) *token2=0;
      len = rcTransformTranslate(token1, token2, dest, writtenchars, TransformLine, MAX_LINE_LENGTH);
    }
    if (!len) break;
    if (len>maxchars) break;
    TransformLine[len] = 0;
    _tcscpy(dest,TransformLine);
    writtenchars = len;
  }
  while (false);
  ASSERT(writtenchars<maxchars);

  return writtenchars;
}



#define VARERR(c) if (--maxchars && --maxchars) \
{ TRACE("VARERR #" #c); *dest++ = ERROR_SIGNAL_CHAR; *dest++ = *_T(#c); break; } else break

//
bool rcVarReplace(TCHAR *src, TCHAR *dest, size_t maxchars, const rcVarLookup *lookup, DWORD lParam)
{
  DWORD v;
  int type;
  bool bOk;
  rcVarPointers ptrs;
  TCHAR *outside;

  ASSERT_WRITEABLE(src);
  ASSERT_WRITEABLE(dest);
  ASSERT_READABLEC(lookup,4);
  ASSERT_READABLEC(lookup,lookup->nSize);

  while (*src && maxchars)
  {
    ASSERT((int)dest>100);
    outside = NULL;
    v = 0;
    switch (*src)
    {
    case OPEN_CHAR:
      type = rcVarType(++src);
      if (type & vtfSpecialBlock)
      {
        /* use a special block parser */ 
        switch (type)
        {
        case vtQuote:
          v = rcVarDoQuote(src, dest, maxchars, outside);
          if (!v) VARERR(Q); /* bad {{ }} quote */ 
          ASSERT(v<maxchars);
          break;

        case vtLiteral:
          v = rcVarDoLiteral(src, dest, maxchars, outside);
          if (!v) VARERR(L); /* bad {[ ]} quote */ 
          ASSERT(v<maxchars);
          break;
        }
      }
      else
      {
        /* use the standard var block parser */ 
        bOk = rcFindMatchingBracket(src, ptrs, outside);
        if (!bOk) type = vtError;

        BeginVarBlock SeparateSections(ptrs);

        switch (type)
        {
        case vtAt: /* {@...} contextual vars */ 
          if (lookup
            && lookup->nSize >= RCVARLOOKUP1_SIZE
            && lookup->At != NULL)
          {
            ++ptrs.NameStart;
            ASSERT_EXECABLE(lookup->At);
            v = lookup->At(&ptrs, dest, maxchars, lParam);
          }
          if (!v) VARERR(@);
          break;

        case vtPlain: /* {...} global vars */ 
          v = rcVarStdPlainLookup(&ptrs, dest, maxchars, lParam);
          if (!v) VARERR(P);
          break;

        case vtBang: VARERR(R); /* {!...} */ 
/*
          if (lookup
            && lookup->nSize >= RCVARLOOKUP1_SIZE
            && lookup->Bang != NULL)
          {
            this needs to do something intelligent
          }
          break;
*/

        case vtDefault: /* {} default value */ 
          if (lookup
            && lookup->nSize >= RCVARLOOKUP1_SIZE
            && lookup->Def != NULL)
          {
            ASSERT_EXECABLE(lookup->Def);
            v = lookup->Def(&ptrs, dest, maxchars, lParam);
          }
          if (!v) VARERR(D);
          break;

          /* user error */ 
        case vtError: VARERR(E);
        case vtReserved: VARERR(R);

          /* code error */ 
        default: VARERR(?);
        }
        ASSERT(v<maxchars);
        if (v)
        {
          v = rcVarApplyTransform(&ptrs, dest, maxchars, v, lParam);
          ASSERT(v<maxchars);
        }
      }
      ASSERT(v<maxchars);
      dest += v;
      maxchars -= v;
      ASSERT(outside!=NULL);
      ASSERT(outside>src);
      if (outside) src = outside;
      ASSERT((int)src>100);
      break;

    case ESCAPE_CHAR:
      v = rcVarEscape(++src, dest, maxchars);
      ASSERT(DEST_INC(v)<maxchars);
      ASSERT_WRITEABLE(dest);
      dest += DEST_INC(v);
      ASSERT_WRITEABLE(dest);
      maxchars -= DEST_INC(v);
      v = SRC_INC(v);
      /* don't let skips go over the end */ 
      while (v--) if (!*++src) break;
      break;

    default:
      ASSERT_WRITEABLE(dest);
      _tccpy(dest, src);
      dest = _tcsinc(dest);
      src = _tcsinc(src);
      ASSERT(maxchars!=0);
      --maxchars;
      ASSERT_WRITEABLE(dest);
    }
    ASSERT((int)dest>100);
  }
  *dest = 0;

  return maxchars!=0;
}
