#ifndef INC_FOWLER_NOLL_VO_HASH_H
#define INC_FOWLER_NOLL_VO_HASH_H

namespace hasher
{
/* Fowler/Noll/Vo hash algorithm (FNV-1a)
 * http://www.isthe.com/chongo/tech/comp/fnv/
 * "The basis of the FNV hash algorithm was taken from an idea sent as
 *  reviewer comments to the IEEE POSIX P1003.2 committee by Glenn Fowler
 *  and Phong Vo. In a subsequent ballot round: Landon Curt Noll improved
 *  on their algorithm."
 *
 *
 * C++ adaptation by RabidCow
 */
  class fnv
  {
  public:
    typedef unsigned hash_t;
    enum { chunk_len = 1 };

  protected:
    hash_t hash;
    enum {
      FNV_prime = 16777619,
      offset_basis = 2166136261
    };

  public:
    fnv(void) : hash(offset_basis) { }

    inline void reset(void) { hash = offset_basis; }
    inline void operator()(char key_i)
    {
      hash ^= key_i;
      hash *= FNV_prime;
    }
    inline void operator()(wchar_t key_i)
    {
      (*this)(static_cast<char>(key_i&0xff));
      (*this)(static_cast<char>(key_i>>8));
    }
    inline void finish(void) { }

    inline operator hash_t(void) const { return hash; }
  };
}

#endif
