#include "module.h"

void PaintWnd(WndInfo *w,HDC hdc)
{
  ASSERT_WRITEABLE(w);

#ifdef NO_SKIN
	Rectangle(hdc,0,0,w->Pos.Width.Value,w->Pos.Height.Value);
#else
  if (!w->emSkin.Bitmap)
  {
    /* pens and brushes don't seem to preserve alpha */ 
    /* so just fill it in by setting the pixels      */ 
    BITMAP bm;
    COLORREF *bits;
    HBITMAP h = (HBITMAP)GetCurrentObject(hdc,OBJ_BITMAP);
    int x,y;
    COLORREF f,l,d,li,di;

    f = w->cSkin.Face.Value;
    switch (w->cSkin.Style)
    {
    case s_EdgeColors::RaisedStyle:
      li = di = f;
      l = w->cSkin.Edge[s_EdgeColors::LightEdge].Value;
      d = w->cSkin.Edge[s_EdgeColors::DarkEdge].Value;
      break;
    case s_EdgeColors::SunkenStyle:
      li = di = f;
      d = w->cSkin.Edge[s_EdgeColors::LightEdge].Value;
      l = w->cSkin.Edge[s_EdgeColors::DarkEdge].Value;
      break;
    case s_EdgeColors::RidgeStyle:
      l = w->cSkin.Edge[s_EdgeColors::LightEdge].Value;
      d = w->cSkin.Edge[s_EdgeColors::DarkEdge].Value;
      li = d; di = l;
      break;
    case s_EdgeColors::GrooveStyle:
      d = w->cSkin.Edge[s_EdgeColors::LightEdge].Value;
      l = w->cSkin.Edge[s_EdgeColors::DarkEdge].Value;
      li = d; di = l;
      break;
    case s_EdgeColors::RaisedButtonStyle:
      li = w->cSkin.Edge[s_EdgeColors::LightEdge].Value;
      di = w->cSkin.Edge[s_EdgeColors::DarkEdge].Value;
      l = w->cSkin.Edge[s_EdgeColors::BrLightEdge].Value;
      d = w->cSkin.Edge[s_EdgeColors::DkDarkEdge].Value;
      break;
    case s_EdgeColors::SunkenButtonStyle:
      di = w->cSkin.Edge[s_EdgeColors::LightEdge].Value;
      l = w->cSkin.Edge[s_EdgeColors::DarkEdge].Value;
      d = w->cSkin.Edge[s_EdgeColors::BrLightEdge].Value;
      li = w->cSkin.Edge[s_EdgeColors::DkDarkEdge].Value;
      break;
    case s_EdgeColors::FlatStyle:
    default:
      l = d = li = di = f;
      break;
    }
    /* need to swap red and green since COLORREFs aren't correct...      */ 
    /* yes, my perser *does* read it correctly, it's just backwards here */ 
#define SWAPRB(n) (((n)&0xff00ff00) | (((n)&0x00ff0000)>>16) | (((n)&0x000000ff)<<16))
    f = SWAPRB(f);
    l = SWAPRB(l);
    d = SWAPRB(d);
    li = SWAPRB(li);
    di = SWAPRB(di);

    if (!GetObject(h,sizeof(BITMAP),&bm)) return;
    if (bm.bmBitsPixel != 32) return;

    bits = (COLORREF *)bm.bmBits;
    if (bm.bmBits == NULL) return;

    *bits++ = l;
    for (x=1;x<bm.bmWidth;++x)
      *bits++ = d;
    *bits++ = l;
    *bits++ = li;
    for (x=4;x<bm.bmWidth;++x)
      *bits++ = di;
    *bits++ = di;
    *bits++ = d;
    for (y=bm.bmHeight;y>4;--y)
    {
      *bits++ = l;
      *bits++ = li;
      for (x=4;x<bm.bmWidth;++x)
        *bits++ = f;
      *bits++ = di;
      *bits++ = d;
    }
    *bits++ = l;
    *bits++ = li;
    for (x=4;x<bm.bmWidth;++x)
      *bits++ = li;
    *bits++ = di;
    *bits++ = d;
    for (x=1;x<bm.bmWidth;++x)
      *bits++ = l;
    *bits++ = d;
  }
  else
  {
    HDC src = CreateCompatibleDC(hdc);
    HGDIOBJ hbmOld = SelectObject(src,w->emSkin.Bitmap);
    RECT rDest;
    SetRect(&rDest,0,0,w->Pos.Width.Value,w->Pos.Height.Value);
    LSTransEdgeBlt(hdc,&rDest,src,&w->emSkin,FALSE);
    SelectObject(src,hbmOld);
    DeleteDC(src);
  }
#endif /* NO_SKIN */ 

  /* draw stuff on top of that here */ 
}

#define MSG(n) MessageBox(NULL,n,ModuleName,MB_OK)

void SetAlphaChannel(HBITMAP h, BYTE value)
{
  BITMAP bm;
  unsigned long count; /* 16gb bitmap maximum */ 
  COLORREF *bits;

  if (!GetObject(h,sizeof(BITMAP),&bm)) return;
  if (bm.bmBitsPixel != 32) return;

  count = bm.bmHeight * bm.bmWidthBytes/sizeof(COLORREF);
  if (!count) count = bm.bmHeight * bm.bmWidth;
  if (!count) return;
  bits = (COLORREF *)bm.bmBits;
  if (bm.bmBits == NULL) return;

  while (count--)
  {
    if ((*bits&0xffffff) == 0xff00ff)
      *bits = 0;
    else
      *bits |= ((DWORD)value)<<24;
    ++bits;
  }
}

void PremultiplyAlpha(HBITMAP h)
{
  BITMAP bm;
  unsigned long count; /* 16gb bitmap maximum */ 
  unsigned r,g,b,a;
  COLORREF *bits;

  if (!GetObject(h,sizeof(BITMAP),&bm)) return;
  if (bm.bmBitsPixel != 32) return;

  count = bm.bmHeight * bm.bmWidthBytes/sizeof(COLORREF);
  if (!count) count = bm.bmHeight * bm.bmWidth;
  if (!count) return;
  ++count;
  bits = (COLORREF *)bm.bmBits;
  if (bm.bmBits == NULL) return;

  /* make sure it's 64-bit aligned */ 
  if ((DWORD)(bits) & 2)
  {
    a = ((*bits)>>24)&255;
    if ((*bits&0xffffff)==RGB(255,0,255))
      a = 0;
    if (a<255)
    {
      r = GetRValue(*bits);
      g = GetGValue(*bits);
      b = GetBValue(*bits);
      *bits = 
        (*bits & 0xff000000)
        | RGB(r*a/255,g*a/255,b*a/255);
    }
    ++bits;
    --count;
  }

  static const __int64 lowbytes = 0x00ff00ff00ff00ff;
  /* mmx loop */ 
  __asm {
    mov edx,bits
    mov ecx,count
    shr ecx,1
    movq mm7,lowbytes
PremultiplyLoop:
    /* actual RGB order doesn't matter   read     write    pipe cyc */ 
    /*                                   01234567 01234567 U V      */ 
    movq mm1,[edx]    /* ARGBARGB -> mm1 -------- -1------ U -    1 */ 
    pxor mm0,mm0      /* ________ -> mm0 0------- 0------- - V    1 */ 
    movq mm3,mm1      /* ARGBARGB -> mm3 -1------ ---3---- U -    1 */ 
    punpcklbw mm1,mm0 /* _A_R_G_B -> mm1 01------ -1------ - V    1 */ 
    punpckhbw mm3,mm0 /* _A_R_G_B -> mm3 0--3---- ---3---- U -    1 */ 
    movq mm0,mm1      /* _A_R_G_B -> mm0 -1------ 0------- - V    1 */ 
    movq mm4,mm3      /* _A_R_G_B -> mm4 ---3---- ----4--- U -    1 */ 
    psrlq mm1,48      /* _______A -> mm1 -1------ -1------ - V    1 */ 
    psrlq mm4,48      /* _______A -> mm4 ----4--- ----4--- U -    1 */ 
    movq mm2,mm1      /* _______A -> mm2 -1------ --2----- - V    1 */ 
    movq mm5,mm4      /* _______A -> mm5 ----4--- -----5-- U -    1 */ 
    psllq mm1,16      /* _____A__ -> mm1 -1------ -1------ - V    1 */ 
    psllq mm4,16      /* _____A__ -> mm4 ----4--- ----4--- U -    1 */ 
    por mm1,mm2       /* _____A_A -> mm1 -12----- -1------ - V    1 */ 
    por mm4,mm5       /* _____A_A -> mm4 ----45-- ----4--- U -    1 */ 
    psllq mm1,16      /* ___A_A__ -> mm1 -1------ -1------ - V    1 */ 
    psllq mm4,16      /* ___A_A__ -> mm4 ----4--- ----4--- U -    1 */ 
    por mm1,mm2       /* ___A_A_A -> mm1 -12----- -1------ - V    1 */ 
    por mm4,mm5       /* ___A_A_A -> mm4 ----45-- ----4--- U -    1 */ 
    psllq mm2,48      /* _A______ -> mm2 --2----- --2----- - V    1 */ 
    psllq mm5,48      /* _A______ -> mm5 -----5-- -----5-- U -    1 */ 
    pmullw mm0,mm1    /* __rrggbb -> mm0 01------ 0------- - V?   3 */ 
    pmullw mm3,mm4    /* __rrggbb -> mm3 ---34--- ---3---- U - +2 3 */ 
    psrlq mm0,8       /* ___rrggb -> mm0 0------- 0------- - V    1 */ 
    psrlq mm3,8       /* ___rrggb -> mm3 ---3---- ---3---- U - +2 1 */ 
    pand mm0,mm7      /* ___r_g_b -> mm0 0------7 0------- - V    1 */ 
    pand mm3,mm7      /* ___r_g_b -> mm0 ---3---7 ---3---- U -    1 */ 
    por mm0,mm2       /* _A_r_g_b -> mm0 0-2----- 0------- - V    1 */ 
    por mm3,mm5       /* _A_r_g_b -> mm5 ---3-5-- ---3---- U -    1 */ 
    add edx,8         /*                          edx      - V    1 */ 
    packuswb mm0,mm3  /* ArgbArgb -> mm0 0--3---- 0------- U -    1 */ 
    movq [edx-8],mm0  /* ArgbArgb -> [edx]                 U - +1   */ 
    loop PremultiplyLoop /*                                U V +1   */ 
    emms
    /* heck, the whole U/V pipeline thing is probly out of date, */ 
    /* but I don't know how else to schedule MMX instructions... */ 
  }
  /* get the last one, since we go by twos */ 
  if (count & 1)
  {
    a = ((*bits)>>24)&255;
    if ((*bits&0xffffff)==RGB(255,0,255))
      a = 0;
    if (a<255)
    {
      r = GetRValue(*bits);
      g = GetGValue(*bits);
      b = GetBValue(*bits);
      *bits = 
        (*bits & 0xff000000)
        | RGB(r*a/255,g*a/255,b*a/255);
    }
    ++bits;
  }
}

HBITMAP Create32BitBitmap(int w,int h)
{
  BITMAPINFOHEADER bmInfo;
	void *bits = NULL;
	ZeroMemory(&bmInfo, sizeof(BITMAPINFOHEADER) );
  // Create a 32 bits depth bitmap
	bmInfo.biSize = sizeof(BITMAPINFOHEADER);
	bmInfo.biWidth = w;
	bmInfo.biHeight = h;
	bmInfo.biPlanes = 1;
	bmInfo.biBitCount = 32;
	bmInfo.biCompression = BI_RGB;

	// create the bitmap
	return CreateDIBSection(
		  NULL,
		  (BITMAPINFO*)&bmInfo,
		  DIB_RGB_COLORS,
		  &bits,
		  NULL,
		  0);
}

void RedrawWnd(WndInfo *w)
{
  ASSERT_WRITEABLE(w);

	HDC hdcScreen,hdcWnd;
	hdcScreen = GetDC(w->hWnd);
	hdcWnd = CreateCompatibleDC(hdcScreen);

	if (w->hbmBackBuffer==NULL)
    w->hbmBackBuffer = Create32BitBitmap(w->Pos.Width.Value,w->Pos.Height.Value);

	HBITMAP hbmOld;
	hbmOld = (HBITMAP)SelectObject(hdcWnd,w->hbmBackBuffer);

  /* If it's a layered window (bTrueAlpha) then we can't  */ 
  /* switch back to using regular paint methods with an   */ 
  /* alpha value of 255 without destroying and recreating */ 
  /* the window, so keep using UpdateLayeredWindow.       */ 

	if ((w->Alpha<255 || w->bTrueAlpha) && dAlphaBlend!=NULL)
	{
		BLENDFUNCTION bl;
		bl.BlendOp = AC_SRC_OVER;
		bl.BlendFlags = 0;
		bl.SourceConstantAlpha = (BYTE)w->Alpha;
    bl.AlphaFormat = w->emSkin.HasAlpha ? AC_SRC_ALPHA : 0;

		if (!w->bTrueAlpha)
		{
			HDC hdcAlpha = CreateCompatibleDC(hdcScreen);
			HBITMAP hbmAlpha = Create32BitBitmap(w->Pos.Width.Value,w->Pos.Height.Value);
			hbmAlpha = (HBITMAP)SelectObject(hdcAlpha,hbmAlpha);

			RECT area;
			GetWindowRect(w->hWnd,&area);
			LSPaintDesktop(hdcWnd,0,0,w->Pos.Width.Value,w->Pos.Height.Value,area.left,area.top);

			PaintWnd(w,hdcAlpha);
      /* doesn't really need to be deselected, */ 
      /* but needs to be the right bitmap      */ 
			hbmAlpha = (HBITMAP)SelectObject(hdcAlpha,hbmAlpha);
      if (bl.AlphaFormat == AC_SRC_ALPHA)
        PremultiplyAlpha(hbmAlpha);
      else
        SetAlphaChannel(hbmAlpha, 255);
			hbmAlpha = (HBITMAP)SelectObject(hdcAlpha,hbmAlpha);

			dAlphaBlend(hdcWnd, 0, 0, w->Pos.Width.Value, w->Pos.Height.Value,
        hdcAlpha, 0, 0, w->Pos.Width.Value, w->Pos.Height.Value, bl);

			hbmAlpha = (HBITMAP)SelectObject(hdcAlpha,hbmAlpha);
			DeleteObject(hbmAlpha);
			DeleteDC(hdcAlpha);

			SelectObject(hdcWnd,hbmOld);
			DeleteDC(hdcWnd);

			InvalidateRect(w->hWnd,NULL,FALSE);
		}
		else
		{
			PaintWnd(w,hdcWnd);
      if (bl.AlphaFormat == AC_SRC_ALPHA)
        PremultiplyAlpha(w->hbmBackBuffer);
      else
        SetAlphaChannel(w->hbmBackBuffer, 255);

			POINT ptSRC = {0,0};
      POINT p = {w->Pos.X.Value,w->Pos.Y.Value};
      SIZE sz = {w->Pos.Width.Value,w->Pos.Height.Value};
			/* not sure why we always do this for layered windows */ 
			/* probably something weird about parents for them    */ 
			LSConvertCoords(w->Pos.Monitor,NULL,NULL,&p.x,&p.y);
			dUpdateLayeredWindow(w->hWnd,NULL,&p,&sz,hdcWnd,&ptSRC,
				RGB(255,0,255),&bl,0|ULW_ALPHA/*|ULW_COLORKEY*/);
      /* COLORKEY + per pixel alpha does not work */ 

			SelectObject(hdcWnd,hbmOld);
			DeleteDC(hdcWnd);
		}
	}
	else
	{
		PaintWnd(w,hdcWnd);

		SelectObject(hdcWnd,hbmOld);
		DeleteDC(hdcWnd);
		InvalidateRect(w->hWnd,NULL,FALSE);
	}

	ReleaseDC(w->hWnd,hdcScreen);
}
