#include "module.h"
#include "droptarget.h"

// create a window and mark it as part of litestep
bool CreateWnd(WndInfo *w)
{
	const static UINT LiteStepMessages[] = {
		LM_GETREVID,
#ifdef LM_MONITORSREFRESHED
    LM_MONITORSREFRESHED,
#endif
		0
	};

	ASSERT_WRITEABLE(w);
  ASSERT_READABLE(w->Name);

  TRACE_MSG(("CreateWnd(%1);",w->Name));

  /* create as on top always, otherwise WS_EX_LAYERED makes position horribly wrong */ 
	w->hWnd =
		CreateWindowEx(
			WS_EX_TOOLWINDOW|WS_EX_TOPMOST|(w->bTrueAlpha?WS_EX_LAYERED:0),
			ModuleName,
			NULL,
			WS_POPUP,
			0, 0, 0, 0, /* size/pos is set later */ 
			NULL, /* no parent for on top windows */ 
			NULL,
			Module.hInstance,
			NULL);
	if (!w->hWnd) return false;

	SetWindowLong(w->hWnd,UGWL_WNDINFO_PTR,(LONG)w);
	SetWindowLong(w->hWnd,GWL_USERDATA,magicDWord);

#ifndef SINGLE_WINDOW
  /* only show up once in revid list */ 
  if (w->pNext != NULL)
	  SendMessage(GetLitestepWnd(), LM_REGISTERMESSAGE,
		  (WPARAM) w->hWnd, (LPARAM) (LiteStepMessages+1));
  else
#endif
	SendMessage(GetLitestepWnd(), LM_REGISTERMESSAGE,
		(WPARAM) w->hWnd, (LPARAM) LiteStepMessages);

  UpdateWndOnTop(w);
  UpdateWndHidden(w);

#ifndef NO_SCRIPT_ACTIONS
  DWORD d;
  w->DropTarget = new SkinBoxDropTarget();
  switch (d = w->DropTarget->Register(w))
  {
  case E_OUTOFMEMORY: MessageBox(w->hWnd, "Out of memory.","SkinBox: DropTarger.Register()",MB_OK); break;
  case S_OK: break;
  case DRAGDROP_E_INVALIDHWND: MessageBox(w->hWnd, "Invalid HWND.","SkinBox: DropTarger.Register()",MB_OK); break;
  case DRAGDROP_E_ALREADYREGISTERED: MessageBox(w->hWnd, "Already registered.","SkinBox: DropTarger.Register()",MB_OK); break;
  default:
    {
      char b[500];
      sprintf(b,"Distracted by flying monkeys.\nIDropTarget: 0x%08X\nHWND: 0x%08X\nSucky: 0x%08X",
        &w->DropTarget,w->hWnd,d);
      MessageBox(w->hWnd, b,"SkinBox: DropTarger.Register()",MB_OK); break;
    }
  }
  ExecAction(w,OnCreate);
#endif

	return true;
}

// update size and position of window
void UpdateWndPos(WndInfo *w)
{
  RECT rMonitor;

	ASSERT_WRITEABLE(w);

  /* should move this out? */ 
  w->Pos.Monitor = LSGetMonitor(w->Pos.MonitorName);

	if (WndInBox(w))
		GetWindowRect(w->hParentWnd,&rMonitor);
	else
		LSGetMonitorScreenRect(w->Pos.Monitor,&rMonitor);

  w->szParent.cx = rMonitor.right - rMonitor.left;
  w->szParent.cy = rMonitor.bottom - rMonitor.top;
  CalcOrd(&w->Pos.X,0,w->szParent.cx);
  CalcOrd(&w->Pos.Y,0,w->szParent.cy);
  CalcOrd(&w->Pos.Width,0,w->szParent.cx);
  CalcOrd(&w->Pos.Height,0,w->szParent.cy);

  {
    TRACE_BLOCK(tBlock);
    TRACE_MSG(("%1 => <%2!5d!, %3!5d!> x <%4!5d!, %5!5d!>",
      w->Name,
      w->Pos.X.Value, w->Pos.Y.Value,
      w->Pos.Width.Value, w->Pos.Height.Value));
    if (w->bTrueAlpha) TRACE("   LAYERED");
    if (w->bOnTop) TRACE("   ONTOP");
  }

  POINT p = {w->Pos.X.Value,w->Pos.Y.Value};
  SIZE sz = {w->Pos.Width.Value,w->Pos.Height.Value};
	/* convert the coords unless we are parented */ 
	if (w->bTrueAlpha && dUpdateLayeredWindow!=NULL)
  {
    /* umm... for some reason this was left out before, but I don't know why */ 
		LSConvertCoords(w->Pos.Monitor,NULL,NULL,&p.x,&p.y);
		dUpdateLayeredWindow(w->hWnd,NULL,&p,&sz,NULL,NULL,0,NULL,0);
  }
	else
  {
    LSConvertCoords(w->Pos.Monitor,
      w->bOnTop?NULL:w->hParentWnd,
      Module.hBottomParentWnd,
      &p.x,&p.y);
		SetWindowPos(w->hWnd,NULL,p.x,p.y,sz.cx,sz.cy,SWP_NOZORDER|SWP_NOACTIVATE);
  }

	DeleteObject(w->hbmBackBuffer);
	w->hbmBackBuffer = NULL;

	RedrawWnd(w);
}

// update on top/bottem setting
void UpdateWndOnTop(WndInfo *w)
{
	ASSERT_READABLE(w);

  /* Remove WS_POPUP and make it a WS_CHILD */ 
	SetWindowLong(w->hWnd, GWL_STYLE,
		(GetWindowLong(w->hWnd, GWL_STYLE) &~ WS_POPUP)|WS_CHILD);

	/* need to reset position because ConvertCoords depends on ontop */ 
	UpdateWndPos(w);

	if (WndInBox(w)) {
		/* in box, set parent to box window                              */ 
		SetParent(w->hWnd, w->hParentWnd);
		/* and leave as WS_CHILD (ignore bOnTop)                         */ 
	} else {
		/* Set the Parent window to Desktop for "onBottom" emulation.    */ 
		/* Or set the Parent to NULL if "onTop" is desired               */ 
		SetParent(w->hWnd, w->bOnTop ? NULL : w->hParentWnd);
		/* Set the Style back to WS_POPUP so things function correctly.  */ 
		SetWindowLong(w->hWnd, GWL_STYLE,
		 (GetWindowLong(w->hWnd, GWL_STYLE) &~ WS_CHILD)|WS_POPUP);
	}
}

// update hide/show
void UpdateWndHidden(WndInfo *w)
{
	ASSERT_READABLE(w);
	ShowWindow(w->hWnd,w->bHidden?SW_HIDE:SW_SHOWNA);
}

#ifndef SINGLE_WINDOW
void CreateWindows(void)
{
	WndInfo *w;

	w = Module.pWnds;
	while (w)
	{
		ASSERT_READABLE(w);
		CreateWnd(w);
		w = w->pNext;
	}
}

void AddWindow(WndInfo *w)
{
	ASSERT_READABLE(w);

	WndInfo **p = &Module.pWnds;
	while (*p) p = &(*p)->pNext;
	*p = w;
	CreateWnd(w);
}
#endif /* SINGLE_WINDOW */ 
